(function (root, factory) {
    root.logStorage = factory(root, root._, root.localStorage, root.hcapPromise);
}(typeof window !== 'undefined' ? window : this, function (global, _, storage, hcapPromise) {
    var NOTIFY_OFFSET = 800;
    function MINUTES(m) { return 1000 * 60 * m; }
    function KB(kb) { return 1024 * kb; }
    function MB(mb) { return 1024 * KB(mb); }
    function makeNotifyEventId(eventId) {
        return eventId + NOTIFY_OFFSET;
    }

    var COLLECT_INTERVAL = MINUTES(30);
    var ITEM_LIMIT_SIZE = KB(10);
    var MAX_STORAGE_SIZE = MB(3);
    var LOG_PREFIX = 'PCD_LOG_';
    var VERSION = '20181015';
    var DEBUG = false;
    var STRING_MAP_KEY = LOG_PREFIX + 'STRING_MAP';
    var stringMap = {};
    var EVENT = {
        USAGE: {
            CHANNEL: 110,
            OTT: 120
        },
        STATUS: {
            POWER_AWAKE: 210,
            POWER_WARM: 211,
            POWER_ON: 212,
            POWER_OFF: 213,
            POWER_REBOOT: 214,
            CRASH_OFF: 221,
            CRASH_REBOOT: 222,
            CPU: 231,
            MEMORY: 232,
            NETWORK_DISCONNECT: 241,
            NETWORK_CONNECT: 242,
            SSID: 243,
            AP_CONNECTION: 244,
            SIGNAL_STRENGTH: 245,
            RSSI_SNR: 246,
            CHECK_IN: 251,
            CHECK_OUT: 252
        },
        IOT: {
            CONNECTED: 301,
            DISCONNECTED: 302,
            CHANGED: 311,
            CONTROLLED: 312
        },
        NOTIFY: {
            CLOSE_ALL: NOTIFY_OFFSET,
            get CLOSE_OTT() { return makeNotifyEventId(EVENT.USAGE.OTT); }
        }
    };
    var TV_INFO = {
        roomId: '',
        tvSn: ''
    };
    var intervalObject = null;
    var storageSize = {};
    var initialized = false;
    var logKeyInfo = {
        status: {
            type: 'status',
            prefix: LOG_PREFIX + 'STATUS_',
            index: 0,
            get currentKey() { return this.prefix + this.index; }
        },
        iot: {
            type: 'iot',
            prefix: LOG_PREFIX + 'IOT_',
            index: 0,
            get currentKey() { return this.prefix + this.index; }
        },
        usage: {
            type: 'monitoring',
            prefix: LOG_PREFIX + 'USAGE_',
            index: 0,
            get currentKey() { return this.prefix + this.index; }
        }
    };

    function log(message) {
        if(DEBUG){
            message = '--- logStorage : ' + message + ' ---';
            global.console.warn(message);
        }
    }

    function upsert(eventId, name, value) {
        if(!isValidEventId(eventId)) {
            return false;
        }

        name = upsertStringMap(name);

        if(needUpdateLastUsageLog(eventId)) {
            lastUsageLog = handleUsageLastLog(eventId);
            if(_.get(lastUsageLog, 'name') == name) {
                return true;
            }
        }

        saveNewLog(getCurrentKey(eventId), createLog(eventId, name, value));
        checkStorageSize();
        return true;
    }

    function isValidEventId(eventId) {
        var validEventIds = [];
        _.forEach(EVENT,  function(value, key) {
            validEventIds = _.concat(validEventIds, _.map(value));
        });
        return _.includes(validEventIds, eventId);
    }

    function needUpdateLastUsageLog(eventId) {
        if(isUsageLog(eventId)) {
            return true;
        }
        else {
            var STATUS = [
                EVENT.STATUS.POWER_WARM,
                EVENT.STATUS.POWER_OFF,
                EVENT.STATUS.POWER_REBOOT
            ];
            return _.includes(_.concat(STATUS, _.map(EVENT.NOTIFY)), eventId);
        }
    }

    function isUsageLog(eventId) {
        return _.includes(_.map(EVENT.USAGE), eventId);
    }

    function handleUsageLastLog(eventId) {
        var usageLastLogsInfo = getLastLogsInfo(logKeyInfo.usage);
        var usageLastLog = _.last(usageLastLogsInfo.logs);

        if(needUpdateToAt(eventId, usageLastLog)) {
            usageLastLog.toAt = Date.now();
            saveLogsToStorage(usageLastLogsInfo.currentKey, usageLastLogsInfo.logs);
        }

        return usageLastLog;
    }

    function getLastLogsInfo(logKey) {
        var lastLogs = getLogs(logKey.currentKey);
        var currentKey = logKey.currentKey;

        if(_.isEmpty(lastLogs) && logKey.index > 0) {
            --logKey.index;
            lastLogs = getLogs(logKey.currentKey);
            currentKey = logKey.currentKey;
            ++logKey.index;
        }

        return {
            logs: lastLogs,
            currentKey: currentKey
        };
    }

    function needUpdateToAt(eventId, usageLastLog) {
        if(!usageLastLog) {
            return false;
        }

        if(isNotifyEvents(eventId) && eventId !== EVENT.NOTIFY.CLOSE_ALL) {
            if((eventId - usageLastLog.eventId) !== NOTIFY_OFFSET) {
                return false;
            }
        }
        return _.isNull(usageLastLog.toAt);
    }

    function isNotifyEvents(eventId) {
        return _.includes(_.map(EVENT.NOTIFY), eventId);
    }

    function getCurrentKey(eventId) {
        /* eventId 1xx, 2xx, 3xx */
        var idx = Number((eventId+"")[0]);
        return logKeyInfo[['', 'usage', 'status', 'iot'][idx]].currentKey;
    }

    function createLog(eventId, name, logValue, fromAt, toAt) {
        var IDX = { EVENT_ID: 0, NAME: 1, VALUE: 2, FROM_AT: 3, TO_AT: 4 };
        return {
            get eventId() { return this.value[IDX.EVENT_ID]; },
            get name() { return this.value[IDX.NAME]; },
            set name(text) { this.value[IDX.NAME] = text; },

            set val(val) { this.value[IDX.VALUE] = val; },

            get fromAt() { return this.value[IDX.FROM_AT]; },
            set fromAt(date) { this.value[IDX.FROM_AT] = date; },

            get toAt() { return this.value[IDX.TO_AT]; },
            set toAt(date) { this.value[IDX.TO_AT] = date; },
            value: [
                eventId,
                name || null,
                logValue || null,
                fromAt || Date.now(),
                toAt || null,
            ]
        };
    }

    function saveNewLog(currentKey, newLog) {
        var logs = _.concat(getLogs(currentKey), newLog);
        saveLogsToStorage(currentKey, logs);
    }

    function saveLogsToStorage(currentKey, logs) {
        var rawLogs = _.map(logs, function(log) {
            return log.value;
        });
        try {
            storage.setItem(currentKey, JSON.stringify(rawLogs));
        }
        catch(e) {
            log('saveLogsToStorage fail: ' + _.get(e, 'message'));
        }
    }

    function sendEvent(eventId, name, value) {
        if(window.COM_TYPE !== "IP" || (isUsageLog(eventId) && storage.getItem('esStatus') !== "1")) {
            return false;
        }
        return upsert(eventId, name, value);
    }

    function collectTvStatus() {
        clearInterval(intervalObject);
        intervalObject = setInterval(collectTvStatusCallback, COLLECT_INTERVAL);
    }

    function collectTvStatusCallback() {
        var collectStatusInfo = getCollectStatusInfo();

        return Promise.all(_.map(collectStatusInfo, function(info) {
                return info.api();
            }))
            .then(function (results) {
                var resultsInfo = zipEventsAndResult(collectStatusInfo, results);
                upsertStatusEvents(resultsInfo);
            })
            .catch(function (e) {
                log('collectTvStatus error: ' + _.get(e, 'message'));
            });
    }

    function getCollectStatusInfo() {
        return [
            makeTvStatus(_.get(hcapPromise, 'system.getCpuUsage', _.noop), [ makeEvents(EVENT.STATUS.CPU, 'percentage') ]),
            makeTvStatus(_.get(hcapPromise, 'system.getMemoryUsage', _.noop), [ makeEvents(EVENT.STATUS.MEMORY, 'percentage') ]),
            makeTvStatus(_.get(hcapPromise, 'network.getNetworkInformation', _.noop), [
                makeEvents(EVENT.STATUS.SSID, 'ssid'),
                makeEvents(EVENT.STATUS.AP_CONNECTION, 'ssid'),
            ]),
            makeTvStatus(_.get(hcapPromise, 'network.getWifiDiagnostics', _.noop), [
                makeEvents(EVENT.STATUS.SIGNAL_STRENGTH, 'rssi'),
                makeEvents(EVENT.STATUS.RSSI_SNR, function(result) {
                    if(result) {
                        var RSSI = getSafeStringToNumber(_.get(result, 'rssi'));
                        var NOISE = getSafeStringToNumber(_.get(result, 'noise'));
                        return RSSI - NOISE;
                    }
                    return result;
                })
            ]),
        ];
    }

    function makeTvStatus(api, events) {
        return {
            api: api,
            events: events
        }
    }

    function makeEvents(eventId, key) {
        return {
            eventId: eventId,
            key: key
        };
    }

    function getSafeStringToNumber(str) {
        if(_.isString(str)) {
            return Number(str.replace(/\D/g, ''));
        }
        return 0;
    }

    function getResultValue(result, key) {
        if(_.isString(key)) {
            return _.get(result, key);
        }
        else {
            return key(result);
        }
    }

    function zipEventsAndResult(collectStatusInfo, results) {
        return _.zipWith(_.map(collectStatusInfo, 'events'), results, function(events, result) {
            return _.map(events, function(event) {
                return {
                    eventId: _.get(event, 'eventId'),
                    value: getResultValue(result, _.get(event, 'key'))
                };
            });
        });
    }
    function upsertStatusEvents(resultsInfo) {
        _.forEach(resultsInfo, function(results) {
            _.forEach(results, function(result) {
                if(isValidValue(result.value)) {
                    upsert(result.eventId, '', result.value);
                }
            });
        });
    }

    function isValidValue(value) {
        return !(_.isUndefined(value) || _.isNull(value) || value === '');
    }

    function sendLogsToServer() {
        log('send STATUS log & monitoring log [START]');

        if(!checkSendData()) {
            return Promise.reject();
        }

        return Promise.mapSeries(getMapLogKeyInfo(), function(info) {
                if(info.key === 'status' || info.key === 'iot' || storage.getItem('esStatus') === "1") {
                    return sendLogToServer(info.logKey)
                        .catch(function(message) {
                            log('sendLogToServer(' + info.key + ') error: ' + message);
                        });
                }
            });
    }

    function checkSendData() {
        if(!initialized) {
            log('not initialized');
            return false;
        }
        // We always should use current room number.
        // TODO: What if room_number is changed? if(TV_INFO.roomId !== storage.getItem('room_number').
        TV_INFO.roomId = storage.getItem('room_number');
        if (_.isEmpty(TV_INFO.roomId)) {
            log('roomId is not valid(' + TV_INFO.roomId + ')');
            return false;
        }
        return true;
    }

    function getMapLogKeyInfo() {
        return _.map(logKeyInfo, function(value, key) {
            return {
                key: key,
                logKey: value
            };
        });
    }

    function sendLogToServer(logKey) {
        var params = {
            data: makeLogData(logKey)
        };
        if(_.isEmpty(params.data)) {
            return Promise.resolve();
        }
        var url = '/api/v2/rooms/' + TV_INFO.roomId + '/tv/' + logKey.type;
        var callBack = makeServerCallBack(logKey);

        return global.requestHandler.requestAPI('POST', url, params, callBack);
    }

    function makeLogData(logKey) {
        var LOG = [ 'eventId', 'name', 'value', 'fromAt', 'toAt' ];
        var stringMapToValues = _.invert(stringMap);

        return _.flattenDeep(_.map(getSortedLogsKeys(logKey), function(key) {
            return _.chain(getLogs(key))
                    .filter(function(log) {
                        return (logKey.type === 'monitoring') ? log.toAt : true;
                    })
                    .map(function(log) {
                        log.name = _.get(stringMapToValues, log.name, null);
                        handleTimeValues(log);
                        var values = _.assign(_.zipObject(LOG, log.value), TV_INFO);
                        return _.pickBy(values, function(val) {
                            return !_.isNull(val);
                        });
                    })
                    .value();
            }));
    }

    function getLogs(key) {
        try {
            var rawLogs = JSON.parse(storage.getItem(key));
            return _.map(rawLogs, function(log) {
                return createLog(log[0], log[1], log[2], log[3], log[4]);
            });
        } catch (e) {
            return [];
        }
    }

    function handleTimeValues(log) {
        if(log.toAt && log.fromAt) {
            log.val =  log.toAt - log.fromAt;
        }
        log.fromAt = getISOString(log.fromAt);
        log.toAt  = getISOString(log.toAt);
    }

    // Do not refacor as belows:
    // new Date(timestamp).toISOString()
    // the error can occur in webOS2.
    function getISOString(timestamp) {
        var ISOString = null;
        var date = null;

        if (timestamp) {
            date = new Date(timestamp);
            ISOString = date.toISOString();
        }
        return ISOString;
    }

    function makeServerCallBack(logKey) {
        return function(response) {
            if(_.get(response, 'status') === 'success') {
                flush(logKey);
            }
        };
    }

    function flush(logKey) {
        var prefix = _.get(logKey, 'prefix', LOG_PREFIX);
        var usageLastLog = null;

        if(_.get(logKey, 'type') === 'usage') {
            var usageLastLogsInfo = getLastLogsInfo(logKeyInfo.usage);
            usageLastLog = _.last(usageLastLogsInfo.logs);
            usageLastLog = (usageLastLog && usageLastLog.toAt) ? null : usageLastLog;
            var newStringMap = _.isNull(usageLastLog) ? {} : _.pickBy(stringMap, function(v){ return v == usageLastLog.name; });
            clearStringMap(newStringMap);
        }

        _.forEach(Object.keys(storage), function (key) {
            if(_.startsWith(key, prefix)) {
                storage.removeItem(key);
            }
        });

        resetIndex(prefix);
        if(usageLastLog) {
            saveNewLog(logKeyInfo.usage.currentKey, usageLastLog)
        }
        updateStorageSize();
    }

    function resetIndex(prefix) {
        prefix = prefix || LOG_PREFIX;
        _.forEach(logKeyInfo, function(logKey) {
            if(_.startsWith(logKey.prefix, prefix)) {
                logKey.index = 0;
            }
        });
    }

    function init(tvRoomId, tvSerialNumber) {
        log('init : '+ tvRoomId + ' / '+ tvSerialNumber);
        if (tvRoomId && tvSerialNumber) {
            checkStorageSize();
            TV_INFO.roomId = tvRoomId;
            TV_INFO.tvSn = tvSerialNumber;
            initialized = true;
            log('init : [SUCCESS]');
            collectTvStatus();
            loadStringsMap();
        }
    }

    function checkStorageSize() {
        updateStorageSize();
        if (storageSize.total > MAX_STORAGE_SIZE) {
            sendLogsToServer();
        }
    }

    function updateStorageSize() {
        var statusSizes = getSizeAndUpdateIndex(logKeyInfo.status);
        var usageSizes = getSizeAndUpdateIndex(logKeyInfo.usage);

        storageSize = {
            statusSizes: statusSizes,
            usageSizes: usageSizes,
            total: _.sum(_.concat(statusSizes, usageSizes))
        };
    }

    function getSizeAndUpdateIndex(logKey) {
        var sizes = getStorageSizeBy(logKey);
        if(_.last(sizes) >= ITEM_LIMIT_SIZE) {
            ++logKey.index;
        }
        return sizes;
    }

    function getStorageSizeBy(logKey) {
        return _.map(getSortedLogsKeys(logKey), function(key) {
            // Javascript uses UTF-16 in localStroage. Thus, we need to multiply 2 about length.
            return ((_.size(storage[key]) + _.size(key)) * 2);
        });
    }

    function getSortedLogsKeys(logKey) {
        return _.chain(Object.keys(storage))
            .filter(function(key) {
                return _.startsWith(key, _.get(logKey, 'prefix'));
            })
            .sortBy(function(key) {
                return _.toNumber(_.last(_.split(key, '_')));
            })
            .value();
    }

    function loadStringsMap() {
        try {
            stringMap = JSON.parse(storage.getItem(STRING_MAP_KEY));
            if(!stringMap) {
                stringMap = {};
            }
        } catch (e) {
        }
    }

    function upsertStringMap(text) {
        if(text && _.isUndefined(stringMap[text])) {
            stringMap[text] = (_.max(_.map(stringMap)) || 0) + 1;
            storage.setItem(STRING_MAP_KEY, JSON.stringify(stringMap));
        }
        return _.get(stringMap, text, null);
    }

    function clearStringMap(map) {
        stringMap = _.isUndefined(map) ? {} : map;
        storage.removeItem(STRING_MAP_KEY);
        storage.setItem(STRING_MAP_KEY, JSON.stringify(stringMap));
    }

    var returnObj = {
        get storageSize() { return storageSize; },
        code: EVENT,
        init: init,
        sendEvent: sendEvent,
        flush: flush,
        sendLogsToServer: sendLogsToServer,
        version: VERSION
    };

    if(typeof chai !== 'undefined' && _.get(chai, 'version')) {
        return _.assign(returnObj, {
            isInitialized: function() {
                return initialized;
            },
            getTVInfo: function() {
                return TV_INFO;
            },
            reset: function() {
                TV_INFO = {
                    roomId: '',
                    tvSn: ''
                };
                storageSize = {};
                initialized = false;
                resetIndex();
                storage.clear();
                clearStringMap();
            },
            hookItemLimitSize: function(newValue) {
                var old = ITEM_LIMIT_SIZE;
                ITEM_LIMIT_SIZE = newValue;
                return old;
            },
            hookMaxStorageSize: function(newValue) {
                var old = MAX_STORAGE_SIZE;
                MAX_STORAGE_SIZE = newValue;
                return old;
            },
            hookSendLogsToServer: function(newValue) {
                var old = sendLogsToServer;
                sendLogsToServer = newValue;
                return old;
            },
            hookServerCallBack: function(newValue) {
                var old = makeServerCallBack;
                makeServerCallBack = function() {
                    return newValue;
                };
                return old;
            },
            hookCollectTvStatus: function(newValue) {
                var old = collectTvStatus;
                collectTvStatus = newValue;
                return old;
            },
            hookCollectTvStatusCallBack: function(newValue) {
                var old = collectTvStatusCallback;
                collectTvStatusCallback = newValue;
                return old;
            },
            hookUpsert: function(newValue) {
                var old = upsert;
                upsert = newValue;
                return old;
            },
            hookHcapPromise: function(newValue) {
                var old = hcapPromise;
                hcapPromise = newValue;
                return old;
            },
            stringMap: stringMap,
            createLog: createLog,
            getLastLogsInfo: getLastLogsInfo,
            getStorageSizeBy: getStorageSizeBy,
            logKeyInfo: logKeyInfo,
            COLLECT_INTERVAL: COLLECT_INTERVAL,
            collectTvStatusCallback: collectTvStatusCallback
        });
    }
    else {
        return returnObj;
    }
}));